
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/poll.h>

#define BUF_LIM 0x100000

/*
 *   The driver will apply POLLIN status after each interrupt
 *   has incremented a global variable. It will execute
 *   wake_up_interruptible at the conclusion of each interrupt
 *   after the variable has been incremented.
 *
 *   This sleeps in poll(), waiting for that new data. It then
 *   reads the new data. If the data is not correct, it writes
 *   a message to standard-error.
 *
 *   A "lost" interrupt will not fool this because any lost
 *   interrupts will not have incremented the variable and
 *   sent a wakeup to poll. This detects the failure of
 *   poll() to return the correct status (for whatever reason).
 */

int main()
{
    char buf[BUF_LIM];
    long long timeval;
    long long localval;
    long lost_events;
    int fd;
    struct pollfd pfd;
    if((fd = open("THING", O_RDONLY)) < 0)
    {
        perror("open");
        exit(EXIT_FAILURE);
    }
    timeval = localval = 0LL;
    if(nice(-19))             // Make high priority
        perror("nice");
    buf[BUF_LIM-1] = 0x00;    // Dirty the otherwise unused buffer
    if(mlockall(MCL_CURRENT)) // Lock whatever we have
        perror("mlockall");
    (void)usleep(10);         // Let kernel do some house-keeping
    lost_events = 0L;
    for(;;)
    {
        pfd.fd = fd;
        pfd.events = POLLIN;
        pfd.revents = 0; 
        poll(&pfd, 1, -1);
        if(pfd.revents & POLLIN)
        {
            if(read(fd, &timeval, sizeof(timeval)) != sizeof(timeval))
                 perror("read");
            else if(++localval != timeval)
            {
                 lost_events += (long) (timeval - localval);
                 fprintf(stderr, "ERROR: (%lld != %lld) Lost events : %ld\n",
                                 timeval, localval, lost_events);
                 localval = timeval; // Re-sync after error
            }
        }
        else
            fprintf(stderr, "ERROR: Bad status, %u\n", pfd.revents);
    }
    return close(fd);
}
